==> spectra_lexer-0.2.1.tar.gz: spectra_lexer-0.2.1/spectra_lexer/file/parser.py <==
import re
from typing import Dict, Tuple

from spectra_lexer.keys import StenoKeys
from spectra_lexer.rules import RuleMap, StenoRule

# Available bracket pairs for parsing rules.
LEFT_BRACKETS = r'\(\['
RIGHT_BRACKETS = r'\)\]'

# Rule substitutions must match a left bracket, one or more non-brackets, and a right bracket.
SUBRULE_RX = re.compile(r'[{}]'.format(LEFT_BRACKETS) +
                        r'[^{0}{1}]+?'.format(LEFT_BRACKETS, RIGHT_BRACKETS) +
                        r'[{}]'.format(RIGHT_BRACKETS))

class StenoRuleDict(Dict[str, StenoRule]):
    """ Class which gets a freshly loaded source dict of raw JSON entries and parses them
        recursively to get a final dict of independent steno rules indexed by internal name. """

    _src_dict: 'RawRulesDictionary'  # Keep the source dict in the instance to avoid passing it everywhere.

    def __init__(self, src_dict:'RawRulesDictionary'):
        """ Top level parsing method. Goes through source JSON dict and parses every entry using mutual recursion. """
        # Unpack rules from all source dictionaries in the given filename list or directory (built-in by default).
        self._src_dict = src_dict
        # Parse all rules from source dictionary into this one, indexed by name.
        # This will parse entries in a semi-arbitrary order, so make sure not to redo any.
        super().__init__()
        for k in self._src_dict.keys():
            if k not in self:
                self._parse(k)

    def _parse(self, k:str) -> None:
        """ Parse a source dictionary rule into a StenoRule object. """
        raw_rule = self._src_dict[k]
        # We have to substitute in the effects of all child rules. These determine the final letters and rulemap.
        letters, rulemap = self._substitute(raw_rule.pattern)
        # The keys must be converted from RTFCRE form into lexer form.
        keys = StenoKeys.parse(raw_rule.keys)
        # Look for key flags, add them as ending rules, and remove them.
        flags = set(raw_rule.flag_str.split("|")) if raw_rule.flag_str else set()
        if flags:
            rulemap.add_key_rules(flags, len(letters), remove_flags=True)
        description = raw_rule.description
        # For now, just include examples as a line after the description joined with commas.
        if raw_rule.example_str:
            description = "{}\n({})".format(description, raw_rule.example_str.replace("|", ", "))
        self[k] = StenoRule(k, keys, letters, flags, description, rulemap)

    def _substitute(self, pattern:str) -> Tuple[str, RuleMap]:
        """
        From a rule's raw YAML pattern string, find all the child rule references in brackets and make a map
        so the format code can break it down again if needed. For those in () brackets, we must substitute
        in their letters as well: (.d)e(.s) -> des. For [] brackets, only add the rulemap entries.

        Only already-finished rules from the results dict can be directly substituted. Any rules that are
        not finished yet will still contain their own child rules in brackets. If we find one of these,
        we have to parse it first in a recursive manner. Circular references will crash the program.
        """
        rulemap = RuleMap()
        m = SUBRULE_RX.search(pattern)
        while m:
            # For every child rule, strip the parentheses to get the dict key (and the letters for () rules).
            rule_str = m.group()
            if rule_str[0] == '(':
                letters = None
                rule_key = rule_str[1:-1]
            else:
                (letters, rule_key) = rule_str[1:-1].split("|", 1)
            # Look up the child rule and parse it if it hasn't been yet. Even if we aren't using its letters,
            # we still need to parse it at this stage so that the correct reference goes in the rulemap.
            if rule_key not in self:
                if rule_key not in self._src_dict:
                    raise KeyError("Illegal reference: {} in {}".format(rule_key, pattern))
                self._parse(rule_key)
            rule = self[rule_key]
            # Add the rule to the map and substitute in the letters if necessary.
            if not letters:
                letters = rule.letters
            rulemap.add_child(rule, m.start(), len(letters))
            pattern = pattern.replace(rule_str, letters)
            m = SUBRULE_RX.search(pattern)
        return pattern, rulemap

==> spectra_lexer-0.3.0.tar.gz: spectra_lexer-0.3.0/spectra_lexer/file/parser.py <==
import re
from typing import Dict, NamedTuple, Tuple

from spectra_lexer.keys import StenoKeys
from spectra_lexer.rules import RuleMap, StenoRule

# Available bracket pairs for parsing rules.
LEFT_BRACKETS = r'\(\['
RIGHT_BRACKETS = r'\)\]'

# Rule substitutions must match a left bracket, one or more non-brackets, and a right bracket.
SUBRULE_RX = re.compile(r'[{}]'.format(LEFT_BRACKETS) +
                        r'[^{0}{1}]+?'.format(LEFT_BRACKETS, RIGHT_BRACKETS) +
                        r'[{}]'.format(RIGHT_BRACKETS))


class RawRule(NamedTuple):
    """ Data structure for raw string fields read from each line in a JSON rules file. """
    keys: str              # RTFCRE formatted series of steno strokes.
    pattern: str           # English text pattern, consisting of raw letters as well as references to other rules.
    flag_str: str = ""     # Optional pipe-delimited series of flags.
    description: str = ""  # Optional description for when the rule is displayed in the GUI.
    example_str: str = ""  # Optional pipe-delimited series of example translations using this rule.


class RawRuleDict(Dict[str, RawRule]):
    """ Class for an unformatted rules dictionary to be saved to/loaded from disk. """
    def __init__(self, src_dict:dict):
        super().__init__({k: RawRule(*v) for (k, v) in src_dict.items()})


class StenoRuleDict(Dict[str, StenoRule]):
    """ Class which takes a source dict of raw JSON rule entries with nested references and parses
        them recursively to get a final dict of independent steno rules indexed by internal name. """

    _src_dict: RawRuleDict  # Keep the source dict in the instance to avoid passing it everywhere.

    def __init__(self, src_dict:dict):
        """ Top level parsing method. Goes through source JSON dict and parses every entry using mutual recursion. """
        # Unpack rules from source dictionary. If the data isn't in namedtuple form, convert it.
        if not isinstance(src_dict, RawRuleDict):
            src_dict = RawRuleDict(src_dict)
        self._src_dict = src_dict
        # Parse all rules from source dictionary into this one, indexed by name.
        # This will parse entries in a semi-arbitrary order, so make sure not to redo any.
        super().__init__()
        for k in self._src_dict.keys():
            if k not in self:
                self._parse(k)

    def _parse(self, k:str) -> None:
        """ Parse a source dictionary rule into a StenoRule object. """
        raw_rule = self._src_dict[k]
        # We have to substitute in the effects of all child rules. These determine the final letters and rulemap.
        letters, rulemap = self._substitute(raw_rule.pattern)
        # The keys must be converted from RTFCRE form into lexer form.
        keys = StenoKeys.parse(raw_rule.keys)
        # Look for key flags, add them as ending rules, and remove them.
        flags = set(raw_rule.flag_str.split("|")) if raw_rule.flag_str else set()
        if flags:
            rulemap.add_key_rules(flags, len(letters), remove_flags=True)
        description = raw_rule.description
        # For now, just include examples as a line after the description joined with commas.
        if raw_rule.example_str:
            description = "{}\n({})".format(description, raw_rule.example_str.replace("|", ", "))
        self[k] = StenoRule(keys, letters, flags, description, rulemap)

    def _substitute(self, pattern:str) -> Tuple[str, RuleMap]:
        """
        From a rule's raw pattern string, find all the child rule references in brackets and make a map
        so the format code can break it down again if needed. For those in () brackets, we must substitute
        in the letters: (.d)e(.s) -> des. For [] brackets, the letters and reference are given separately.

        Only already-finished rules from the results dict can be directly substituted. Any rules that are
        not finished yet will still contain their own child rules in brackets. If we find one of these,
        we have to parse it first in a recursive manner. Circular references will crash the program.
        """
        rulemap = RuleMap()
        m = SUBRULE_RX.search(pattern)
        while m:
            # For every child rule, strip the parentheses to get the dict key (and the letters for [] rules).
            rule_str = m.group()
            if rule_str[0] == '(':
                letters = None
                rule_key = rule_str[1:-1]
            else:
                (letters, rule_key) = rule_str[1:-1].split("|", 1)
            # Look up the child rule and parse it if it hasn't been yet. Even if we aren't using its letters,
            # we still need to parse it at this stage so that the correct reference goes in the rulemap.
            if rule_key not in self:
                if rule_key not in self._src_dict:
                    raise KeyError("Illegal reference: {} in {}".format(rule_key, pattern))
                self._parse(rule_key)
            rule = self[rule_key]
            # Add the rule to the map and substitute in the letters if necessary.
            if not letters:
                letters = rule.letters
            rulemap.add_child(rule, m.start(), len(letters))
            pattern = pattern.replace(rule_str, letters)
            m = SUBRULE_RX.search(pattern)
        return pattern, rulemap
